import sys
import textwrap
import base64

from IPython.display import Image

from ipywidgets.widgets import FloatProgress, Layout

from hdijupyterutils.guid import ObjectWithGuid

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.sparklogger import SparkLog
from sparkmagic.utils.sparkevents import SparkEvents
from sparkmagic.utils.constants import (
    MAGICS_LOGGER_NAME,
    FINAL_STATEMENT_STATUS,
    MIMETYPE_IMAGE_PNG,
    MIMETYPE_TEXT_HTML,
    MIMETYPE_TEXT_PLAIN,
    COMMAND_INTERRUPTED_MSG,
    COMMAND_CANCELLATION_FAILED_MSG,
)
from .exceptions import (
    LivyUnexpectedStatusException,
    SparkStatementCancelledException,
    SparkStatementCancellationFailedException,
)


class Command(ObjectWithGuid):
    def __init__(self, code, spark_events=None):
        super(Command, self).__init__()
        self.code = textwrap.dedent(code)
        self.logger = SparkLog("Command")
        if spark_events is None:
            spark_events = SparkEvents()
        self._spark_events = spark_events

    def __repr__(self):
        return "Command({}, ...)".format(repr(self.code))

    def __eq__(self, other):
        return self.code == other.code

    def __ne__(self, other):
        return not self == other

    def execute(self, session):
        self._spark_events.emit_statement_execution_start_event(
            session.guid, session.kind, session.id, self.guid
        )
        statement_id = -1
        try:
            session.wait_for_idle()
            data = {"code": self.code}
            response = session.http_client.post_statement(session.id, data)
            statement_id = response["id"]
            output = self._get_statement_output(session, statement_id)
        except KeyboardInterrupt as e:
            self._spark_events.emit_statement_execution_end_event(
                session.guid,
                session.kind,
                session.id,
                self.guid,
                statement_id,
                False,
                e.__class__.__name__,
                str(e),
            )
            try:
                if statement_id >= 0:
                    response = session.http_client.cancel_statement(
                        session.id, statement_id
                    )
                    session.wait_for_idle()
            except:
                raise SparkStatementCancellationFailedException(
                    COMMAND_CANCELLATION_FAILED_MSG
                )
            else:
                raise SparkStatementCancelledException(COMMAND_INTERRUPTED_MSG)
        except Exception as e:
            self._spark_events.emit_statement_execution_end_event(
                session.guid,
                session.kind,
                session.id,
                self.guid,
                statement_id,
                False,
                e.__class__.__name__,
                str(e),
            )
            raise
        else:
            self._spark_events.emit_statement_execution_end_event(
                session.guid,
                session.kind,
                session.id,
                self.guid,
                statement_id,
                True,
                "",
                "",
            )
            return output

    def _get_statement_output(self, session, statement_id):
        retries = 1
        progress = FloatProgress(
            value=0.0,
            min=0,
            max=1.0,
            step=0.01,
            description="Progress:",
            bar_style="info",
            orientation="horizontal",
            layout=Layout(width="50%", height="25px"),
        )
        session.ipython_display.display(progress)

        while True:
            statement = session.http_client.get_statement(session.id, statement_id)
            status = statement["state"].lower()

            self.logger.debug(
                "Status of statement {} is {}.".format(statement_id, status)
            )

            if status not in FINAL_STATEMENT_STATUS:
                progress.value = statement.get("progress", 0.0)
                session.sleep(retries)
                retries += 1
            else:
                statement_output = statement["output"]
                progress.close()

                if statement_output is None:
                    return (True, "", MIMETYPE_TEXT_PLAIN)

                if statement_output["status"] == "ok":
                    data = statement_output["data"]
                    if MIMETYPE_IMAGE_PNG in data:
                        image = Image(base64.b64decode(data[MIMETYPE_IMAGE_PNG]))
                        return (True, image, MIMETYPE_IMAGE_PNG)
                    elif MIMETYPE_TEXT_HTML in data:
                        return (True, data[MIMETYPE_TEXT_HTML], MIMETYPE_TEXT_HTML)
                    else:
                        return (True, data[MIMETYPE_TEXT_PLAIN], MIMETYPE_TEXT_PLAIN)
                elif statement_output["status"] == "error":
                    return (
                        False,
                        statement_output["evalue"]
                        + "\n"
                        + "".join(statement_output["traceback"]),
                        MIMETYPE_TEXT_PLAIN,
                    )
                else:
                    raise LivyUnexpectedStatusException(
                        "Unknown output status from Livy: '{}'".format(
                            statement_output["status"]
                        )
                    )



# Distributed under the terms of the Modified BSD License.
import threading
from time import sleep, time

from hdijupyterutils.guid import ObjectWithGuid

import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants
from sparkmagic.utils.sparklogger import SparkLog
from sparkmagic.utils.sparkevents import SparkEvents
from sparkmagic.utils.utils import get_sessions_info_html
from .configurableretrypolicy import ConfigurableRetryPolicy
from .command import Command
from .exceptions import (
    LivyClientTimeoutException,
    LivyUnexpectedStatusException,
    BadUserDataException,
    SqlContextNotFoundException,
)


class _HeartbeatThread(threading.Thread):
    def __init__(self, livy_session, refresh_seconds, retry_seconds, run_at_most=None):
        """
        Parameters
        ----------
        livy_session : LivySession
        refresh_seconds: int
            The seconds to sleep between refreshing the Livy session status and info
        retry_seconds: int
            The seconds to sleep before retrying on a failed refresh
        run_at_most: int, optional
            The max number of loops to execute before ending this thread
        """
        super(_HeartbeatThread, self).__init__()

        self.livy_session = livy_session
        self.refresh_seconds = refresh_seconds
        self.retry_seconds = retry_seconds
        if run_at_most is None:
            # a billion iterations should be sufficient
            run_at_most = int(1e9)
        self.run_at_most = run_at_most

    def run(self):
        loop_counter = 0
        if self.livy_session is None:
            print("Will not start heartbeat thread because self.livy_session is None")
            return

        self.livy_session.logger.info(
            "Starting heartbeat for session {}".format(self.livy_session.id)
        )

        while self.livy_session is not None and loop_counter < self.run_at_most:
            loop_counter += 1

            try:
                sleep_time = self.refresh_seconds
                self.livy_session.refresh_status_and_info()
            except Exception as e:
                sleep_time = self.retry_seconds
                # The built-in python logger has exception handling built in. If you expose
                # the "exception" function in the SparkLog class then you could just make this
                # self.livy_session.logger.exception("some useful message") and it'll print
                # out the stack trace too.
                self.livy_session.logger.error("{}".format(e))

            sleep(sleep_time)

    def stop(self):
        if self.livy_session is not None:
            self.livy_session.logger.info(
                "Stopping heartbeat for session {}".format(self.livy_session.id)
            )

        self.livy_session = None
        self.join()


class LivySession(ObjectWithGuid):
    def __init__(
        self,
        http_client,
        properties,
        ipython_display,
        session_id=-1,
        spark_events=None,
        heartbeat_timeout=0,
        heartbeat_thread=None,
    ):
        super(LivySession, self).__init__()
        assert constants.LIVY_KIND_PARAM in list(properties.keys())
        kind = properties[constants.LIVY_KIND_PARAM]

        should_heartbeat = False
        if heartbeat_timeout > 0:
            should_heartbeat = True
            properties[constants.LIVY_HEARTBEAT_TIMEOUT_PARAM] = heartbeat_timeout
        elif constants.LIVY_HEARTBEAT_TIMEOUT_PARAM in list(properties.keys()):
            properties.pop(constants.LIVY_HEARTBEAT_TIMEOUT_PARAM)

        self.properties = properties
        self.ipython_display = ipython_display
        self._should_heartbeat = should_heartbeat
        self._user_passed_heartbeat_thread = heartbeat_thread

        if spark_events is None:
            spark_events = SparkEvents()
        self._spark_events = spark_events

        self._policy = ConfigurableRetryPolicy(
            retry_seconds_to_sleep_list=[0.2, 0.5, 0.5, 1, 1, 2], max_retries=5000
        )
        wait_for_idle_timeout_seconds = conf.wait_for_idle_timeout_seconds()

        assert wait_for_idle_timeout_seconds > 0

        self.logger = SparkLog("LivySession")

        kind = kind.lower()
        if kind not in constants.SESSION_KINDS_SUPPORTED:
            raise BadUserDataException(
                "Session of kind '{}' not supported. Session must be of kinds {}.".format(
                    kind, ", ".join(constants.SESSION_KINDS_SUPPORTED)
                )
            )

        self._app_id = None
        self._user = None
        self._logs = ""
        self._http_client = http_client
        self._wait_for_idle_timeout_seconds = wait_for_idle_timeout_seconds
        self._printed_resource_warning = False

        self.kind = kind
        self.id = session_id
        self.session_info = ""

        self._heartbeat_thread = None
        if session_id == -1:
            self.status = constants.NOT_STARTED_SESSION_STATUS
        else:
            self.status = constants.BUSY_SESSION_STATUS
            self._start_heartbeat_thread()

    def __str__(self):
        return "Session id: {}\tYARN id: {}\tKind: {}\tState: {}\n\tSpark UI: {}\n\tDriver Log: {}".format(
            self.id,
            self.get_app_id(),
            self.kind,
            self.status,
            self.get_spark_ui_url(),
            self.get_driver_log_url(),
        )

    def start(self):
        """Start the session against actual livy server."""
        self._spark_events.emit_session_creation_start_event(self.guid, self.kind)
        self._printed_resource_warning = False

        try:
            r = self._http_client.post_session(self.properties)
            self.id = r["id"]
            self.status = str(r["state"])

            self.ipython_display.writeln("Starting Spark application")

            # Start heartbeat thread to keep Livy interactive session alive.
            self._start_heartbeat_thread()

            # We wait for livy_session_startup_timeout_seconds() for the session to start up.
            try:
                self.wait_for_idle(conf.livy_session_startup_timeout_seconds())
            except LivyClientTimeoutException:
                raise LivyClientTimeoutException(
                    "Session {} did not start up in {} seconds.".format(
                        self.id, conf.livy_session_startup_timeout_seconds()
                    )
                )

            html = get_sessions_info_html([self], self.id)
            self.ipython_display.html(html)

            command = Command("spark")
            (success, out, mimetype) = command.execute(self)

            if success:
                self.ipython_display.writeln("SparkSession available as 'spark'.")
                self.sql_context_variable_name = "spark"
            else:
                command = Command("sqlContext")
                (success, out, mimetype) = command.execute(self)
                if success:
                    self.ipython_display.writeln("SparkContext available as 'sc'.")
                    if "hive" in out.lower():
                        self.ipython_display.writeln(
                            "HiveContext available as 'sqlContext'."
                        )
                    else:
                        self.ipython_display.writeln(
                            "SqlContext available as 'sqlContext'."
                        )
                    self.sql_context_variable_name = "sqlContext"
                else:
                    raise SqlContextNotFoundException(
                        "Neither SparkSession nor HiveContext/SqlContext is available."
                    )
        except Exception as e:
            self._spark_events.emit_session_creation_end_event(
                self.guid,
                self.kind,
                self.id,
                self.status,
                False,
                e.__class__.__name__,
                str(e),
            )
            raise
        else:
            self._spark_events.emit_session_creation_end_event(
                self.guid, self.kind, self.id, self.status, True, "", ""
            )

    def get_app_id(self):
        if self._app_id is None:
            self._app_id = self._http_client.get_session(self.id).get("appId")
        return self._app_id

    def get_app_info(self):
        appInfo = self._http_client.get_session(self.id).get("appInfo")
        return appInfo if appInfo is not None else {}

    def get_app_info_member(self, member_name):
        return self.get_app_info().get(member_name)

    def get_driver_log_url(self):
        return self.get_app_info_member("driverLogUrl")

    def get_logs(self):
        log_array = self._http_client.get_all_session_logs(self.id)["log"]
        self._logs = "\n".join(log_array)
        return self._logs

    def get_spark_ui_url(self):
        return self.get_app_info_member("sparkUiUrl")

    def get_user(self):
        if self._user is None:
            session = self._http_client.get_session(self.id)
            self._user = session.get("proxyUser", session.get("owner"))
        return self._user

    @property
    def http_client(self):
        return self._http_client

    @property
    def endpoint(self):
        return self._http_client.endpoint

    @staticmethod
    def is_final_status(status):
        return status in constants.FINAL_STATUS

    def is_posted(self):
        return self.status != constants.NOT_STARTED_SESSION_STATUS

    def delete(self):
        session_id = self.id
        self._spark_events.emit_session_deletion_start_event(
            self.guid, self.kind, session_id, self.status
        )

        try:
            self.logger.debug("Deleting session '{}'".format(session_id))

            if self.status != constants.NOT_STARTED_SESSION_STATUS:
                self._http_client.delete_session(session_id)
                self._stop_heartbeat_thread()
                self.status = constants.DEAD_SESSION_STATUS
                self.id = -1
            else:
                self.ipython_display.send_error(
                    "Cannot delete session {} that is in state '{}'.".format(
                        session_id, self.status
                    )
                )

        except Exception as e:
            self._spark_events.emit_session_deletion_end_event(
                self.guid,
                self.kind,
                session_id,
                self.status,
                False,
                e.__class__.__name__,
                str(e),
            )
            raise
        else:
            self._spark_events.emit_session_deletion_end_event(
                self.guid, self.kind, session_id, self.status, True, "", ""
            )

    def wait_for_idle(self, seconds_to_wait=None):
        """Wait for session to go to idle status. Sleep meanwhile.

        Parameters:
            seconds_to_wait : number of seconds to wait before giving up.
        """
        if seconds_to_wait is None:
            seconds_to_wait = self._wait_for_idle_timeout_seconds

        retries = 1
        while True:
            self.refresh_status_and_info()
            if self.status == constants.IDLE_SESSION_STATUS:
                return

            if self.status in constants.FINAL_STATUS:
                error = "Session {} unexpectedly reached final status '{}'.".format(
                    self.id, self.status
                )
                self.logger.error(error)
                raise LivyUnexpectedStatusException(
                    "{} See logs:\n{}".format(error, self.get_logs())
                )

            if seconds_to_wait <= 0.0:
                error = "Session {} did not reach idle status in time. Current status is {}.".format(
                    self.id, self.status
                )
                self.logger.error(error)
                raise LivyClientTimeoutException(error)

            if (
                constants.YARN_RESOURCE_LIMIT_MSG in self.session_info
                and not self._printed_resource_warning
            ):
                self.ipython_display.send_error(
                    constants.RESOURCE_LIMIT_WARNING.format(
                        conf.resource_limit_mitigation_suggestion()
                    )
                )
                self._printed_resource_warning = True

            start_time = time()
            sleep_time = self._policy.seconds_to_sleep(retries)
            retries += 1

            self.logger.debug(
                "Session {} in state {}. Sleeping {} seconds.".format(
                    self.id, self.status, sleep_time
                )
            )
            sleep(sleep_time)
            seconds_to_wait -= time() - start_time

    def sleep(self, retries):
        sleep(self._policy.seconds_to_sleep(retries))

    # This function will refresh the status and get the logs in a single call.
    # Only the status will be returned as the return value.
    def refresh_status_and_info(self):
        response = self._http_client.get_session(self.id)
        status = response["state"]
        log_array = response["log"]

        if status in constants.POSSIBLE_SESSION_STATUS:
            self.status = status
            self.session_info = "\n".join(log_array)
        else:
            raise LivyUnexpectedStatusException(
                "Status '{}' not supported by session.".format(status)
            )

    def _start_heartbeat_thread(self):
        if self._should_heartbeat and self._heartbeat_thread is None:
            refresh_seconds = conf.heartbeat_refresh_seconds()
            retry_seconds = conf.heartbeat_retry_seconds()

            if self._user_passed_heartbeat_thread is None:
                self._heartbeat_thread = _HeartbeatThread(
                    self, refresh_seconds, retry_seconds
                )
            else:
                self._heartbeat_thread = self._user_passed_heartbeat_thread

            self._heartbeat_thread.daemon = True
            self._heartbeat_thread.start()

    def _stop_heartbeat_thread(self):
        if self._heartbeat_thread is not None:
            self._heartbeat_thread.stop()
            self._heartbeat_thread = None

    def get_row_html(self, current_session_id):
        return """<tr><td>{0}</td><td>{1}</td><td>{2}</td><td>{3}</td><td>{4}</td><td>{5}</td><td>{6}</td><td>{7}</td></tr>""".format(
            self.id,
            self.get_app_id(),
            self.kind,
            self.status,
            self.get_html_link("Link", self.get_spark_ui_url()),
            self.get_html_link("Link", self.get_driver_log_url()),
            self.get_user(),
            ""
            if current_session_id is None or current_session_id != self.id
            else "\u2714",
        )

    @staticmethod
    def get_html_link(text, url):
        if url is not None:
            return """<a target="_blank" href="{1}">{0}</a>""".format(text, url)
        else:
            return ""



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants
from sparkmagic.utils.sparklogger import SparkLog
from .sessionmanager import SessionManager
from .livyreliablehttpclient import LivyReliableHttpClient
from .livysession import LivySession
from sparkmagic.utils.constants import MAGICS_LOGGER_NAME


class SparkController(object):
    def __init__(self, ipython_display):
        self.logger = SparkLog("SparkController")
        self.ipython_display = ipython_display
        self.session_manager = SessionManager(ipython_display)
        # this is to reuse the already created http clients
        # since the reliablehttpclient uses requests session
        self._http_clients = {}

    def get_app_id(self, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return session_to_use.get_app_id()

    def get_driver_log_url(self, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return session_to_use.get_driver_log_url()

    def get_logs(self, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return session_to_use.get_logs()

    def get_spark_ui_url(self, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return session_to_use.get_spark_ui_url()

    def run_command(self, command, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return command.execute(session_to_use)

    def run_sqlquery(self, sqlquery, client_name=None):
        session_to_use = self.get_session_by_name_or_default(client_name)
        return sqlquery.execute(session_to_use)

    def get_all_sessions_endpoint(self, endpoint):
        http_client = self._http_client(endpoint)
        sessions = http_client.get_sessions()["sessions"]
        supported_sessions = filter(
            lambda s: (
                s[constants.LIVY_KIND_PARAM] in constants.SESSION_KINDS_SUPPORTED
            ),
            sessions,
        )
        session_list = [
            self._livy_session(
                http_client,
                {constants.LIVY_KIND_PARAM: s[constants.LIVY_KIND_PARAM]},
                self.ipython_display,
                s["id"],
            )
            for s in supported_sessions
        ]
        for s in session_list:
            s.refresh_status_and_info()
        return session_list

    def get_all_sessions_endpoint_info(self, endpoint):
        sessions = self.get_all_sessions_endpoint(endpoint)
        return [str(s) for s in sessions]

    def cleanup(self):
        self.session_manager.clean_up_all()

    def cleanup_endpoint(self, endpoint):
        for session in self.get_all_sessions_endpoint(endpoint):
            session.delete()

    def delete_session_by_name(self, name):
        self.session_manager.delete_client(name)

    def delete_session_by_id(self, endpoint, session_id):
        name = self.session_manager.get_session_name_by_id_endpoint(
            session_id, endpoint
        )

        if name in self.session_manager.get_sessions_list():
            self.delete_session_by_name(name)
        else:
            http_client = self._http_client(endpoint)
            response = http_client.get_session(session_id)
            http_client = self._http_client(endpoint)
            session = self._livy_session(
                http_client,
                {constants.LIVY_KIND_PARAM: response[constants.LIVY_KIND_PARAM]},
                self.ipython_display,
                session_id,
            )
            session.delete()

    def add_session(self, name, endpoint, skip_if_exists, properties):
        if skip_if_exists and (name in self.session_manager.get_sessions_list()):
            self.logger.debug(
                "Skipping {} because it already exists in list of sessions.".format(
                    name
                )
            )
            return
        http_client = self._http_client(endpoint)
        session = self._livy_session(http_client, properties, self.ipython_display)

        try:
            session.start()
        except:
            if session.is_posted():
                session.delete()
            raise
        else:
            self.session_manager.add_session(name, session)

    def get_session_id_for_client(self, name):
        return self.session_manager.get_session_id_for_client(name)

    def get_client_keys(self):
        return self.session_manager.get_sessions_list()

    def get_manager_sessions_str(self):
        return self.session_manager.get_sessions_info()

    def get_session_by_name_or_default(self, client_name):
        if client_name is None:
            return self.session_manager.get_any_session()
        else:
            return self.session_manager.get_session(client_name)

    def get_managed_clients(self):
        return self.session_manager.sessions

    @staticmethod
    def _livy_session(http_client, properties, ipython_display, session_id=-1):
        return LivySession(
            http_client,
            properties,
            ipython_display,
            session_id,
            heartbeat_timeout=conf.livy_server_heartbeat_timeout_seconds(),
        )

    def _http_client(self, endpoint):
        if endpoint not in self._http_clients:
            self._http_clients[endpoint] = LivyReliableHttpClient.from_endpoint(
                endpoint
            )
        return self._http_clients[endpoint]



from nose.tools import assert_equals

from sparkmagic.livyclientlib.configurableretrypolicy import ConfigurableRetryPolicy
import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException


def test_with_empty_list():
    times = []
    max_retries = 5
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert_equals(5, policy.seconds_to_sleep(0))
    assert_equals(5, policy.seconds_to_sleep(4))
    assert_equals(5, policy.seconds_to_sleep(5))
    assert_equals(5, policy.seconds_to_sleep(6))

    # Check based on retry count
    assert_equals(True, policy.should_retry(500, False, 0))
    assert_equals(True, policy.should_retry(500, False, 4))
    assert_equals(True, policy.should_retry(500, False, 5))
    assert_equals(False, policy.should_retry(500, False, 6))

    # Check based on status code
    assert_equals(False, policy.should_retry(201, False, 0))
    assert_equals(False, policy.should_retry(201, False, 6))

    # Check based on error
    assert_equals(True, policy.should_retry(201, True, 0))
    assert_equals(True, policy.should_retry(201, True, 6))


def test_with_one_element_list():
    times = [2]
    max_retries = 5
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert_equals(2, policy.seconds_to_sleep(0))
    assert_equals(2, policy.seconds_to_sleep(4))
    assert_equals(2, policy.seconds_to_sleep(5))
    assert_equals(2, policy.seconds_to_sleep(6))

    # Check based on retry count
    assert_equals(True, policy.should_retry(500, False, 0))
    assert_equals(True, policy.should_retry(500, False, 4))
    assert_equals(True, policy.should_retry(500, False, 5))
    assert_equals(False, policy.should_retry(500, False, 6))

    # Check based on status code
    assert_equals(False, policy.should_retry(201, False, 0))
    assert_equals(False, policy.should_retry(201, False, 6))

    # Check based on error
    assert_equals(True, policy.should_retry(201, True, 0))
    assert_equals(True, policy.should_retry(201, True, 6))


def test_with_default_values():
    times = conf.retry_seconds_to_sleep_list()
    max_retries = conf.configurable_retry_policy_max_retries()
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert_equals(times[0], policy.seconds_to_sleep(0))
    assert_equals(times[0], policy.seconds_to_sleep(1))
    assert_equals(times[1], policy.seconds_to_sleep(2))
    assert_equals(times[2], policy.seconds_to_sleep(3))
    assert_equals(times[3], policy.seconds_to_sleep(4))
    assert_equals(times[4], policy.seconds_to_sleep(5))
    assert_equals(times[4], policy.seconds_to_sleep(6))
    assert_equals(times[4], policy.seconds_to_sleep(7))
    assert_equals(times[4], policy.seconds_to_sleep(8))
    assert_equals(times[4], policy.seconds_to_sleep(9))

    # Check based on retry count
    assert_equals(True, policy.should_retry(500, False, 0))
    assert_equals(True, policy.should_retry(500, False, 7))
    assert_equals(True, policy.should_retry(500, False, 8))
    assert_equals(False, policy.should_retry(500, False, 9))

    # Check based on status code
    assert_equals(False, policy.should_retry(201, False, 0))
    assert_equals(False, policy.should_retry(201, False, 9))

    # Check based on error
    assert_equals(True, policy.should_retry(201, True, 0))
    assert_equals(True, policy.should_retry(201, True, 9))


def test_with_negative_values():
    times = [0.1, -1]
    max_retries = 5

    try:
        policy = ConfigurableRetryPolicy(times, max_retries)
        assert False
    except BadUserConfigurationException:
        assert True



# -*- coding: utf-8 -*-

"""Runs Scala, PySpark and SQL statement through Spark using a REST endpoint in remote cluster.
Provides the %spark magic."""

# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
from six import string_types

from IPython.core.magic import Magics, magics_class
from hdijupyterutils.ipythondisplay import IpythonDisplay
from collections import namedtuple
import sparkmagic.utils.configuration as conf
from sparkmagic.utils.sparklogger import SparkLog
from sparkmagic.utils.sparkevents import SparkEvents
from sparkmagic.utils.utils import get_sessions_info_html
from sparkmagic.utils.constants import MIMETYPE_TEXT_HTML
from sparkmagic.livyclientlib.sparkcontroller import SparkController
from sparkmagic.livyclientlib.sqlquery import SQLQuery
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.sparkstorecommand import SparkStoreCommand
from sparkmagic.livyclientlib.exceptions import SparkStatementException
from sparkmagic.livyclientlib.sendpandasdftosparkcommand import (
    SendPandasDfToSparkCommand,
)
from sparkmagic.livyclientlib.sendstringtosparkcommand import SendStringToSparkCommand
from sparkmagic.livyclientlib.exceptions import BadUserDataException

# How to display different cell content types in IPython
SparkOutputHandler = namedtuple("SparkOutputHandler", ["html", "text", "default"])


@magics_class
class SparkMagicBase(Magics):

    _STRING_VAR_TYPE = "str"
    _PANDAS_DATAFRAME_VAR_TYPE = "df"
    _ALLOWED_LOCAL_TO_SPARK_TYPES = [_STRING_VAR_TYPE, _PANDAS_DATAFRAME_VAR_TYPE]

    def __init__(self, shell, data=None, spark_events=None):
        # You must call the parent constructor
        super(SparkMagicBase, self).__init__(shell)

        self.logger = SparkLog("SparkMagics")
        self.ipython_display = IpythonDisplay()
        self.spark_controller = SparkController(self.ipython_display)

        self.logger.debug("Initialized spark magics.")

        if spark_events is None:
            spark_events = SparkEvents()
        spark_events.emit_library_loaded_event()

    def do_send_to_spark(
        self,
        cell,
        input_variable_name,
        var_type,
        output_variable_name,
        max_rows,
        session_name,
    ):
        try:
            input_variable_value = self.shell.user_ns[input_variable_name]
        except KeyError:
            raise BadUserDataException(
                "Variable named {} not found.".format(input_variable_name)
            )
        if input_variable_value is None:
            raise BadUserDataException(
                "Value of {} is None!".format(input_variable_name)
            )

        if not output_variable_name:
            output_variable_name = input_variable_name

        if not max_rows:
            max_rows = conf.default_maxrows()

        input_variable_type = var_type.lower()
        if input_variable_type == self._STRING_VAR_TYPE:
            command = SendStringToSparkCommand(
                input_variable_name, input_variable_value, output_variable_name
            )
        elif input_variable_type == self._PANDAS_DATAFRAME_VAR_TYPE:
            command = SendPandasDfToSparkCommand(
                input_variable_name,
                input_variable_value,
                output_variable_name,
                max_rows,
            )
        else:
            raise BadUserDataException(
                "Invalid or incorrect -t type. Available are: [{}]".format(
                    ",".join(self._ALLOWED_LOCAL_TO_SPARK_TYPES)
                )
            )

        (success, result, mime_type) = self.spark_controller.run_command(command, None)
        if not success:
            self.ipython_display.send_error(result)
        else:
            self.ipython_display.write(
                "Successfully passed '{}' as '{}' to Spark"
                " kernel".format(input_variable_name, output_variable_name)
            )

    def execute_spark(
        self,
        cell,
        output_var,
        samplemethod,
        maxrows,
        samplefraction,
        session_name,
        coerce,
        output_handler=None,
    ):
        output_handler = output_handler or SparkOutputHandler(
            html=self.ipython_display.html,
            text=self.ipython_display.write,
            default=self.ipython_display.display,
        )

        (success, out, mimetype) = self.spark_controller.run_command(
            Command(cell), session_name
        )
        if not success:
            if conf.shutdown_session_on_spark_statement_errors():
                self.spark_controller.cleanup()

            raise SparkStatementException(out)
        else:
            if isinstance(out, string_types):
                if mimetype == MIMETYPE_TEXT_HTML:
                    output_handler.html(out)
                else:
                    output_handler.text(out)
            else:
                output_handler.default(out)
            if output_var is not None:
                spark_store_command = self._spark_store_command(
                    output_var, samplemethod, maxrows, samplefraction, coerce
                )
                df = self.spark_controller.run_command(
                    spark_store_command, session_name
                )
                self.shell.user_ns[output_var] = df

    @staticmethod
    def _spark_store_command(output_var, samplemethod, maxrows, samplefraction, coerce):
        return SparkStoreCommand(
            output_var, samplemethod, maxrows, samplefraction, coerce=coerce
        )

    def execute_sqlquery(
        self,
        cell,
        samplemethod,
        maxrows,
        samplefraction,
        session,
        output_var,
        quiet,
        coerce,
    ):
        sqlquery = self._sqlquery(cell, samplemethod, maxrows, samplefraction, coerce)
        df = self.spark_controller.run_sqlquery(sqlquery, session)
        if output_var is not None:
            self.shell.user_ns[output_var] = df
        if quiet:
            return None
        else:
            return df

    @staticmethod
    def _sqlquery(cell, samplemethod, maxrows, samplefraction, coerce):
        return SQLQuery(cell, samplemethod, maxrows, samplefraction, coerce=coerce)

    def _print_endpoint_info(self, info_sessions, current_session_id):
        if info_sessions:
            info_sessions = sorted(info_sessions, key=lambda s: s.id)
            html = get_sessions_info_html(info_sessions, current_session_id)
            self.ipython_display.html(html)
        else:
            self.ipython_display.html("No active sessions.")



from hdijupyterutils.guid import ObjectWithGuid

from sparkmagic.utils.utils import (
    coerce_pandas_df_to_numeric_datetime,
    records_to_dataframe,
)
import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants
from sparkmagic.utils.sparkevents import SparkEvents
from .command import Command
from .exceptions import DataFrameParseException, BadUserDataException


class SQLQuery(ObjectWithGuid):
    def __init__(
        self,
        query,
        samplemethod=None,
        maxrows=None,
        samplefraction=None,
        spark_events=None,
        coerce=None,
    ):
        super(SQLQuery, self).__init__()

        if samplemethod is None:
            samplemethod = conf.default_samplemethod()
        if maxrows is None:
            maxrows = conf.default_maxrows()
        if samplefraction is None:
            samplefraction = conf.default_samplefraction()

        if samplemethod not in {"take", "sample"}:
            raise BadUserDataException(
                "samplemethod (-m) must be one of (take, sample)"
            )
        if not isinstance(maxrows, int):
            raise BadUserDataException("maxrows (-n) must be an integer")
        if not 0.0 <= samplefraction <= 1.0:
            raise BadUserDataException(
                "samplefraction (-r) must be a float between 0.0 and 1.0"
            )

        self.query = query
        self.samplemethod = samplemethod
        self.maxrows = maxrows
        self.samplefraction = samplefraction
        if spark_events is None:
            spark_events = SparkEvents()
        self._spark_events = spark_events
        self._coerce = coerce

    def to_command(self, kind, sql_context_variable_name):
        if kind == constants.SESSION_KIND_PYSPARK:
            return self._pyspark_command(sql_context_variable_name)
        elif kind == constants.SESSION_KIND_SPARK:
            return self._scala_command(sql_context_variable_name)
        elif kind == constants.SESSION_KIND_SPARKR:
            return self._r_command(sql_context_variable_name)
        else:
            raise BadUserDataException("Kind '{}' is not supported.".format(kind))

    def execute(self, session):
        self._spark_events.emit_sql_execution_start_event(
            session.guid,
            session.kind,
            session.id,
            self.guid,
            self.samplemethod,
            self.maxrows,
            self.samplefraction,
        )
        command_guid = ""
        try:
            command = self.to_command(session.kind, session.sql_context_variable_name)
            command_guid = command.guid
            (success, records_text, mimetype) = command.execute(session)
            if not success:
                raise BadUserDataException(records_text)
            result = records_to_dataframe(records_text, session.kind, self._coerce)
        except Exception as e:
            self._spark_events.emit_sql_execution_end_event(
                session.guid,
                session.kind,
                session.id,
                self.guid,
                command_guid,
                False,
                e.__class__.__name__,
                str(e),
            )
            raise

        else:
            self._spark_events.emit_sql_execution_end_event(
                session.guid,
                session.kind,
                session.id,
                self.guid,
                command_guid,
                True,
                "",
                "",
            )
            return result

    def _pyspark_command(self, sql_context_variable_name):
        # use_unicode=False means the result will be UTF-8-encoded bytes, so we
        # set it to False for Python 2.
        command = '{}.sql(u"""{} """).toJSON(use_unicode=(sys.version_info.major > 2))'.format(
            sql_context_variable_name, self.query
        )
        if self.samplemethod == "sample":
            command = "{}.sample(False, {})".format(command, self.samplefraction)
        if self.maxrows >= 0:
            command = "{}.take({})".format(command, self.maxrows)
        else:
            command = "{}.collect()".format(command)
        print_command = constants.LONG_RANDOM_VARIABLE_NAME
        command = "import sys\nfor {} in {}: print({})".format(
            constants.LONG_RANDOM_VARIABLE_NAME, command, print_command
        )
        return Command(command)

    def _scala_command(self, sql_context_variable_name):
        command = '{}.sql("""{}""").toJSON'.format(
            sql_context_variable_name, self.query
        )
        if self.samplemethod == "sample":
            command = "{}.sample(false, {})".format(command, self.samplefraction)
        if self.maxrows >= 0:
            command = "{}.take({})".format(command, self.maxrows)
        else:
            command = "{}.collect".format(command)
        return Command("{}.foreach(println)".format(command))

    def _r_command(self, sql_context_variable_name):
        if sql_context_variable_name == "spark":
            command = 'sql("{}")'.format(self.query)
        else:
            command = 'sql({}, "{}")'.format(sql_context_variable_name, self.query)
        if self.samplemethod == "sample":
            command = "sample({}, FALSE, {})".format(command, self.samplefraction)
        if self.maxrows >= 0:
            command = "take({},{})".format(command, self.maxrows)
        else:
            command = "collect({})".format(command)
        command = "jsonlite:::toJSON({})".format(command)
        command = "for ({} in ({})) {{cat({})}}".format(
            constants.LONG_RANDOM_VARIABLE_NAME,
            command,
            constants.LONG_RANDOM_VARIABLE_NAME,
        )
        return Command(command)

    # Used only for unit testing
    def __eq__(self, other):
        return (
            self.query == other.query
            and self.samplemethod == other.samplemethod
            and self.maxrows == other.maxrows
            and self.samplefraction == other.samplefraction
            and self._coerce == other._coerce
        )

    def __ne__(self, other):
        return not (self == other)



# Distributed under the terms of the Modified BSD License.
from IPython.core.error import UsageError
from IPython.core.magic_arguments import parse_argstring
import numpy as np
import pandas as pd
import json
import importlib
from collections import OrderedDict

import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants
from sparkmagic.livyclientlib.exceptions import (
    BadUserDataException,
    DataFrameParseException,
    BadUserConfigurationException,
)


def get_coerce_value(coerce):
    if coerce is not None:
        coerce = coerce.lower() in ("yes", "true", "t", "y", "1")
    return coerce


def parse_argstring_or_throw(magic_func, argstring, parse_argstring=parse_argstring):
    """An alternative to the parse_argstring method from IPython.core.magic_arguments.
    Catches IPython.core.error.UsageError and propagates it as a
    livyclientlib.exceptions.BadUserDataException."""
    try:
        return parse_argstring(magic_func, argstring)
    except UsageError as e:
        raise BadUserDataException(str(e))


def coerce_pandas_df_to_numeric_datetime(df):
    for column_name in df.columns:
        coerced = False

        if df[column_name].isnull().all():
            continue

        if not coerced and df[column_name].dtype == np.dtype("object"):
            try:
                df[column_name] = pd.to_datetime(df[column_name], errors="raise")
                coerced = True
            except (ValueError, TypeError, OverflowError):
                pass

        if not coerced and df[column_name].dtype == np.dtype("object"):
            try:
                df[column_name] = pd.to_numeric(df[column_name], errors="raise")
                coerced = True
            except (ValueError, TypeError):
                pass


def records_to_dataframe(records_text, kind, coerce=None):
    if records_text in ["", "[]"]:
        strings = []
    else:
        strings = records_text.strip().split("\n")
    try:
        data_array = [
            json.JSONDecoder(object_pairs_hook=OrderedDict).decode(s) for s in strings
        ]

        if kind == constants.SESSION_KIND_SPARKR and len(data_array) > 0:
            data_array = data_array[0]

        df = pd.DataFrame(data_array)

        if len(data_array) > 0:
            # This will assign the columns in the right order. If we simply did
            # df = pd.DataFrame(data_array, columns=data_array[0].keys())
            # in the code defining df, above, we could get an issue where the first element
            # has some columns as null, and thus would drop the columns from the df altogether.
            # Refer to https://github.com/jupyter-incubator/sparkmagic/issues/346 for
            # more details.
            for data in data_array:
                if len(data.keys()) == len(df.columns):
                    df = df[list(data.keys())]
                    break

        if coerce is None:
            coerce = conf.coerce_dataframe()
        if coerce:
            coerce_pandas_df_to_numeric_datetime(df)

        return df
    except ValueError:
        raise DataFrameParseException(
            "Cannot parse object as JSON: '{}'".format(strings)
        )


def get_sessions_info_html(info_sessions, current_session_id):
    html = (
        """<table>
<tr><th>ID</th><th>YARN Application ID</th><th>Kind</th><th>State</th><th>Spark UI</th><th>Driver log</th><th>User</th><th>Current session?</th></tr>"""
        + "".join(
            [session.get_row_html(current_session_id) for session in info_sessions]
        )
        + "</table>"
    )

    return html


def initialize_auth(args):
    """Creates an authenticatior class instance for the given auth type

    Args:
        args (IPython.core.magics.namespace): The namespace object that is created from
        parsing %spark magic command

    Returns:
        An instance of a valid Authenticator or None if args.auth is 'None'

    Raises:
        sparkmagic.livyclientlib.BadUserConfigurationException: if args.auth is not a valid
        authenticator class.
    """
    if args.auth is None:
        auth = conf.get_auth_value(args.user, args.password)
    else:
        auth = args.auth
    if auth == constants.NO_AUTH:
        return None
    else:
        full_class = conf.authenticators().get(auth)
        if full_class is None:
            raise BadUserConfigurationException("Auth '{}' not supported".format(auth))
        module, class_name = (full_class).rsplit(".", 1)
        events_handler_module = importlib.import_module(module)
        auth_class = getattr(events_handler_module, class_name)
        return auth_class(args)


class Namespace:
    """Namespace to initialize authenticator class with"""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import json

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.constants import LANG_SCALA, LANG_PYTHON
from sparkmagic.controllerwidget.abstractmenuwidget import AbstractMenuWidget


class CreateSessionWidget(AbstractMenuWidget):
    def __init__(
        self,
        spark_controller,
        ipywidget_factory,
        ipython_display,
        endpoints_dropdown_widget,
        refresh_method,
    ):
        # This is nested
        super(CreateSessionWidget, self).__init__(
            spark_controller, ipywidget_factory, ipython_display, True
        )

        self.refresh_method = refresh_method

        self.endpoints_dropdown_widget = endpoints_dropdown_widget

        self.session_widget = self.ipywidget_factory.get_text(
            description="Name:", value="session-name"
        )
        self.lang_widget = self.ipywidget_factory.get_toggle_buttons(
            description="Language:",
            options=[LANG_SCALA, LANG_PYTHON],
        )
        self.properties = self.ipywidget_factory.get_text(
            description="Properties:", value=json.dumps(conf.session_configs())
        )
        self.submit_widget = self.ipywidget_factory.get_submit_button(
            description="Create Session"
        )

        self.children = [
            self.ipywidget_factory.get_html(value="<br/>", width="600px"),
            self.endpoints_dropdown_widget,
            self.session_widget,
            self.lang_widget,
            self.properties,
            self.ipywidget_factory.get_html(value="<br/>", width="600px"),
            self.submit_widget,
        ]

        for child in self.children:
            child.parent_widget = self

    def run(self):
        try:
            properties_json = self.properties.value
            if properties_json.strip() != "":
                conf.override(
                    conf.session_configs.__name__, json.loads(self.properties.value)
                )
        except ValueError as e:
            self.ipython_display.send_error(
                "Session properties must be a valid JSON string. Error:\n{}".format(e)
            )
            return

        endpoint = self.endpoints_dropdown_widget.value
        language = self.lang_widget.value
        alias = self.session_widget.value
        skip = False
        properties = conf.get_session_properties(language)

        try:
            self.spark_controller.add_session(alias, endpoint, skip, properties)
        except ValueError as e:
            self.ipython_display.send_error(
                """Could not add session with
name:
    {}
properties:
    {}

due to error: '{}'""".format(
                    alias, properties, e
                )
            )
            return

        self.refresh_method()
